class TextStyle:
    def __init__(self, text_type):
        self.text_type = text_type 
        self.init()

    def init(self):
        self.size = 12
        self.angle = 0
        self._anchor = 'start'
        self.angle_anchor = 'start'
        self.baseline = 'bottom'
        self.base_offset = 0
        self.format = lambda x: x

    @property
    def anchor(self):
        return self._anchor if self.angle == 0 else self.angle_anchor

    @anchor.setter
    def anchor(self, anchor):
        self._anchor = anchor

    @property
    def css(self):
        return f'''
            .{self.text_type}_label {{
                font-size: {self.size}px;
                text-anchor: {self.anchor};
                dominant-baseline: {self.baseline};
            }}
        '''



from gweld.libs.util_lib import calculate_y_scale, calculate_chart_size, calculate_margins
from gweld.libs.svg_lib import add_tag, add_text
from gweld import VisElement

class AxisElement(VisElement):
    def plot(self, tree, vis):
        margins = calculate_margins(vis)
        size = calculate_chart_size(vis, margins)
        size = (size[0]+1, size[1]+1)

        y_scale = calculate_y_scale(vis.data, vis.style.y_axis_tick_number)

        inner_tree = add_tag(tree, 'svg', attributes={
            'viewBox': f'0 0 {size[0]}, {size[1]}',
            'x': str(margins[0]-1),
            'y': str(margins[1]),
            'width': str(size[0]),
            'height': str(size[1])
        })

        for i in range(1, len(y_scale)):
            y_pos = size[1] - i * size[1]/(len(y_scale)-1)
            if vis.style.show_grid_lines:
                add_tag(inner_tree, 'line', attributes={
                    'x1': '0',
                    'x2': str(size[0]),
                    'y1': str(y_pos),
                    'y2': str(y_pos),
                    'class': 'grid_lines'
                })
        
        # Axes
        
        add_tag(inner_tree, 'line', attributes={
            'x1': str(0),
            'x2': str(0),
            'y1': str(size[1]),
            'y2': str(0),
            'class': 'axis'
        })
        
        add_tag(inner_tree, 'line', attributes={
            'x1': str(0),
            'x2': str(size[0]),
            'y1': str(size[1]),
            'y2': str(size[1]),
            'class': 'axis'
        })
        
        for i, label in enumerate(y_scale):
            y_pos = margins[0] + size[1] - i * size[1]/(len(y_scale)-1)
            add_text(tree, (margins[0]-5, y_pos), str(label), vis.style.text_styles['y_axis'])



from gweld.data import Data
from gweld.data_set import DataSet

from gweld.style.text_style import TextStyle
from gweld.style.circle_text_style import CircleTextStyle
from gweld.style.style import Style

from gweld.elements.vis_element import VisElement
from gweld.elements.text_element import TextElement
from gweld.elements.axis_element import AxisElement
from gweld.elements.bg_element import BGElement

from gweld.charts.chart import Chart
from gweld.charts.bar import Bar
from gweld.charts.pie import Pie
from gweld.charts.line import Line

from gweld.vis import Vis



from gweld import Data

class DataSet(Data):
    def __iadd__(self, other):
        if isinstance(other, Data):
            self.data.append(other)
        else:
            raise TypeError

        return self

    def __len__(self):
        return max([len(data) for data in self.data])

    @property
    def max(self):
        return max([max(data) for data in self.data])

    @property
    def min(self):
        return min([min(data) for data in self.data])



from gweld.libs.util_lib import calculate_y_scale, calculate_chart_size, calculate_margins
from gweld.libs.svg_lib import root_tag, add_tag, add_text
from gweld import Chart, DataSet, Data, AxisElement, BGElement
import math

class Line(Chart):
    def init(self):
        self.elements = [BGElement(), AxisElement()]

    def plot(self, tree, vis):
        if len(vis.data) == 0:
            return tree

        margins = calculate_margins(vis)
        size = calculate_chart_size(vis, margins)

        y_scale = calculate_y_scale(vis.data, vis.style.y_axis_tick_number)

        inner_tree = add_tag(tree, 'svg', attributes={
            'viewBox': f'0 0 {size[0]}, {size[1]}',
            'x': str(margins[0]),
            'y': str(margins[1]),
            'width': str(size[0]),
            'height': str(size[1])
        })

        if isinstance(vis.data, DataSet):
            for i, data in enumerate(vis.data):
                self._plot_line(vis, data, y_scale[-1], size[0], size[1], vis.style.data_colours[i%len(data)], inner_tree)
        elif isinstance(vis.data, Data):
            self._plot_line(vis, vis.data, y_scale[-1], size[0], size[1], vis.style.data_colour, inner_tree)
        else:
            raise Exception
        
        if vis.data.labels:
            for i, label in enumerate(vis.data.labels):
                if not i % vis.style.x_axis_interval:
                    width_per_node = size[0] / (len(vis.data)-1)
                    x = margins[0] + i * width_per_node

                    label_y = margins[1] + size[1] + vis.style.text_styles["x_axis"].size/2
                    add_text(tree, (x, label_y), str(label), vis.style.text_styles['x_axis'])

        return tree

    def _plot_line(self, vis, data, max_value, plot_width, plot_height, colour, tree):
        width_per_node = plot_width / (len(data)-1)
        points = []

        for i, item in enumerate(data):
            x = i * width_per_node
            y = plot_height - (item / max_value) * plot_height
            points.append((x, y))

            if vis.style.show_values == 'all' or (vis.style.show_values == 'limits' and
                    (item == data.max or item == data.min)):
                label_y = y - vis.style.text_styles["value"].size/2

                add_text(tree, (x, label_y), str(item), vis.style.text_styles['value'])

        add_tag(tree, 'polyline', attributes={
            'points': ' '.join([f'{str(point[0])}, {str(point[1])}' for point in points]),
            'fill': 'none',
            'stroke': colour,
            'class': 'line_chart_data_line'
        })



from gweld.libs.util_lib import calculate_y_scale, calculate_chart_size, calculate_margins
from gweld.libs.svg_lib import root_tag, add_tag, add_text
from gweld import Chart, AxisElement, BGElement
import math

class Bar(Chart):
    def init(self):
        self.elements = [BGElement(), AxisElement()]

    def plot(self, tree, vis):
        if len(vis.data) == 0:
            return tree

        margins = calculate_margins(vis)
        size = calculate_chart_size(vis, margins)

        y_scale = calculate_y_scale(vis.data, vis.style.y_axis_tick_number)

        inner_tree = add_tag(tree, 'svg', attributes={
            'viewBox': f'0 0 {size[0]}, {size[1]}',
            'x': str(margins[0]),
            'y': str(margins[1]),
            'width': str(size[0]),
            'height': str(size[1])
        })

        width_per_bar = size[0] / len(vis.data)
        bar_width = width_per_bar * vis.style.bar_width

        for i, item in enumerate(vis.data):
            height = (item / y_scale[-1]) * size[1]
            centre_x = i * width_per_bar + width_per_bar/2

            add_tag(inner_tree, 'rect', attributes={
                'width': str(bar_width),
                'height': str(height),
                'x': str(centre_x - bar_width/2),
                'y': str(size[1] - height),
                'class': 'data_colour'
            })

            if vis.style.show_values == 'all' or (vis.style.show_values == 'limits' and
                    (item == vis.data.max or item == vis.data.min)):
                label_y = size[1] - height - vis.style.text_styles["value"].size/2
                add_text(tree, (margins[0] + centre_x, margins[1] + label_y), str(item), vis.style.text_styles['value'])

            if vis.data.labels:
                if not i % vis.style.x_axis_interval:
                    label_y = size[1] + vis.style.text_styles["x_axis"].size/2
                    add_text(tree, (margins[0] + centre_x, margins[1] + label_y), str(vis.data.labels[i]), vis.style.text_styles['x_axis'])

        return tree

    def _calculate_y_scale(self, data, tick_count=5):
        # Algorithm from: https://stackoverflow.com/a/326746
        lower_bound = 0 # No support for negative numbers.... yet
        upper_bound = data.max

        if upper_bound == 0:
            upper_bound = 1

        data_range = upper_bound - lower_bound
        coarse_tick_size = data_range / (tick_count-1)
        magnitude = math.ceil(math.log(coarse_tick_size, 10)) # Yay floating point arithmetic!

        for tick_size in [0.1, 0.125, 0.15, 0.2, 0.25, 0.3, 0.4, 0.5, 0.6, 0.7, 0.75, 0.8, 0.9, 1]:
            if coarse_tick_size/10**magnitude <= tick_size:
                tick_size *= 10**magnitude
                break

        if not tick_size % 1:
            tick_size = int(tick_size)

        scale = [0]
        while tick_count-1 > 0:
            scale.append(round(scale[-1]+tick_size, 3))

            if scale[-1] > upper_bound:
                break
            tick_count -= 1

        return scale



from gweld.libs.svg_lib import add_tag
from gweld import VisElement

class BGElement(VisElement):
    def plot(self, tree, vis):
        add_tag(tree, 'rect', attributes={
            'x': '0',
            'y': '0',
            'width': str(vis.style.width),
            'height': str(vis.style.height),
            'fill': vis.style.background_colour
        })



from gweld import TextStyle

class CircleTextStyle(TextStyle):
    @property
    def css(self):
        return f'''
            .{self.text_type}_label_upper_left {{
                font-size: {self.size}px;
                text-anchor: {self.anchor if self.anchor == 'middle' else 'start'};
                dominant-baseline: {self.baseline if self.baseline == 'middle' else 'auto'};
            }}

            .{self.text_type}_label_upper_right {{
                font-size: {self.size}px;
                text-anchor: {self.anchor if self.anchor == 'middle' else 'end'};
                dominant-baseline: {self.baseline if self.baseline == 'middle' else 'auto'};
            }}

            .{self.text_type}_label_lower_left {{
                font-size: {self.size}px;
                text-anchor: {self.anchor if self.anchor == 'middle' else 'start'};
                dominant-baseline: {self.baseline if self.baseline == 'middle' else 'hanging'};
            }}

            .{self.text_type}_label_lower_right {{
                font-size: {self.size}px;
                text-anchor: {self.anchor if self.anchor == 'middle' else 'end'};
                dominant-baseline: {self.baseline if self.baseline == 'middle' else 'hanging'};
            }}
        '''



"""
vis.py
======

Module containing the core visualisation object.
"""

from gweld import Data, Chart, Bar, VisElement, Style
from gweld.libs.svg_lib import add_tag, root_tag, to_string

class Vis:
    """
    A visualisation object. The core object worked with to create visualisations.
    """
    def __init__(self, data=Data(), chart=Bar(), elements=[], style=Style()):
        self.init()

        self.data = data
        self.chart = chart
        self.elements += elements
        self.style = style

    def init(self):
        self.elements = []

    def __repr__(self):
        return f'Chart(data={self.data!r}, chart={self.chart!r}, style={self.style!r})'

    def __add__(self, other):
        if isinstance(other, Data):
            return Vis(data=other, chart=self.chart, elements=self.elements, style=self.style)
        if isinstance(other, Chart):
            return Vis(data=self.data, chart=other, elements=self.elements, style=self.style)
        if isinstance(other, VisElement):
            return Vis(data=self.data, chart=self.chart, elements=self.elements+[other], style=self.style)
        if isinstance(other, Style):
            return Vis(data=self.data, chart=self.chart, elements=self.elements, style=other)
        raise TypeError

    def plot(self):
        tree = root_tag(self.style.width, self.style.height)
        add_tag(tree, 'style', text=self.style.css)

        for element in self.chart.elements:
            element.plot(tree, self)
        
        self.chart.plot(tree, self)

        for element in self.elements:
            element.plot(tree, self)

        return to_string(tree)



from lxml import etree

def root_tag(width, height):
    return etree.Element('svg', {'width': str(width), 'height': str(height), 'xmlns': 'http://www.w3.org/2000/svg'})

def add_tag(parent, tag, attributes=None, text=None):
    tag = etree.SubElement(parent, tag, attributes)
    tag.text = str(text)
    return tag 

def add_text(parent, pos, text, style, class_suffix=''):
    add_tag(parent, 'text', attributes={
        'x': str(pos[0]),
        'y': str(pos[1] + style.base_offset),
        'transform': f'rotate({style.angle} {pos[0]} {pos[1]})',
        'class': f'{style.text_type}_label{class_suffix}'
    }, text=style.format(text))

def to_string(tree):
    return etree.tostring(tree, pretty_print=True)



from gweld.libs.svg_lib import root_tag, add_tag, add_text
from gweld import Chart, BGElement
import math

class Pie(Chart):
    def init(self):
        self.elements = [BGElement()]

    def plot(self, tree, vis):
        if len(vis.data) == 0:
            return tree

        # Left, Right
        plot_x = (
            vis.style.width * vis.style.margin[0],
            vis.style.width * vis.style.margin[2]
        )
        # Up, Down
        plot_y = (
            vis.style.height * vis.style.margin[1],
            vis.style.height * vis.style.margin[3]
        )

        plot_width = vis.style.width - plot_x[0] - plot_x[1]
        plot_height = vis.style.height - plot_y[0] - plot_y[1]


        # Plot pie

        angles = self._calculate_angles(vis.data)
        centre = (plot_x[0] + plot_width/2, plot_y[0] + plot_height/2)
        radius = min(plot_width, plot_height)/2 * 0.9
        value_radius = radius * vis.style.pie_value_radius

        # SVG paths can't draw full circle. Backtrack to a circle if we
        # get too close
        if len(vis.data) == 1 or max(angles) > 1.99 * math.pi:
            add_tag(tree, 'circle', attributes={
                'cx': str(centre[0]),
                'cy': str(centre[1]),
                'r': str(radius),
                'fill': vis.style.data_colours[0]
            })
        else:
            x = centre[0]
            y = centre[1] - radius
            angle = 0

            for i, item in enumerate(vis.data):
                (old_x, old_y) = (x,y)
                angle += angles[i]
                x = centre[0] + radius * math.sin(angle)
                y = centre[1] + -(radius * math.cos(angle))
                big = 1 if angles[i] > math.pi else 0

                add_tag(tree, 'path', attributes={
                    'd': f'M{centre[0]},{centre[1]} L{old_x},{old_y} A{radius},{radius} 0 {big},1 {x},{y} Z',
                    'fill': vis.style.data_colours[(i%len(vis.data)) % len(vis.style.data_colours)]
                }, text=str((angles[i], angle)))

            angle = 0
            for i, item in enumerate(vis.data):
                angle += angles[i] / 2
                x = centre[0] + value_radius * math.sin(angle)
                y = centre[1] + -(value_radius * math.cos(angle))

                css_class = '_' 
                if angle < math.pi/2 or angle > math.pi*6/4:
                    css_class += 'upper'
                else:
                    css_class += 'lower'
                if angle < math.pi:
                    css_class += '_left'
                else:
                    css_class += '_right'

                add_text(
                    tree,
                    (x,y),
                    str(item),
                    vis.style.text_styles['circle_value'],
                    css_class
                )

                angle += angles[i] / 2

        add_tag(tree, 'circle', attributes={
            'cx': str(centre[0]),
            'cy': str(centre[1]),
            'r': str(radius * vis.style.pie_inner_radius),
            'fill': vis.style.background_colour
        })

        return tree

    def _calculate_angles(self, data):
        angles = []

        for item in data:
            angles.append(2 * math.pi * (item / sum(data)))

        return angles



class Data:
    def __init__(self, data=[], labels=[]):
        self.data = list(data)
        self.labels = labels

    def __repr__(self):
        return f'Data({self.data!r})'

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]

    @property
    def max(self):
        if len(self.data):
            return max(self.data)
        else:
            return 1

    @property
    def min(self):
        if len(self.data):
            return min(self.data)
        else:
            return 0



from gweld.libs.svg_lib import add_text
from gweld import VisElement

class TextElement(VisElement):
    def __init__(self, text, pos, style):
        self.text = text
        self.pos = pos
        self.style = style

    def plot(self, tree, vis):
        x = self.pos[0] + margins[0]
        y = self.pos[1] + margins[1]

        add_text(tree, (x, y), self.text, self.style)



import math

def calculate_y_scale(data, tick_count=5):
    # Algorithm from: https://stackoverflow.com/a/326746
    lower_bound = 0 # No support for negative numbers.... yet
    upper_bound = data.max
    data_range = upper_bound - lower_bound
    coarse_tick_size = data_range / (tick_count-1)
    magnitude = math.ceil(math.log(coarse_tick_size, 10)) # Yay floating point arithmetic!

    for tick_size in [0.1, 0.125, 0.15, 0.2, 0.25, 0.3, 0.4, 0.5, 0.6, 0.7, 0.75, 0.8, 0.9, 1]:
        if coarse_tick_size/10**magnitude <= tick_size:
            tick_size *= 10**magnitude
            break

    if not tick_size % 1:
        tick_size = int(tick_size)

    scale = [0]
    while tick_count-1 > 0:
        scale.append(round(scale[-1]+tick_size, 3))

        if scale[-1] > upper_bound:
            break
        tick_count -= 1

    return scale

def calculate_margins(vis):
    # Left, Top, Right, Bottom
    return (
        vis.style.width * vis.style.margin[0],
        vis.style.width * vis.style.margin[2],
        vis.style.height * vis.style.margin[1],
        vis.style.height * vis.style.margin[3]
    )

def calculate_chart_size(vis, margins):
    width = vis.style.width - margins[0] - margins[2]
    height = vis.style.height - margins[1] - margins[3]

    return (width, height)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='gweld',
    version='0.0.3',
    author='Ieuan Jones',
    author_email='ieujon@googlemail.com',
    description='A simple charting library',
    long_description=long_description,
    url='https://github.com/ieuanjones/gweld',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.6',
    install_requires=['lxml']
)



from gweld import TextStyle, CircleTextStyle

class Style:
    def __init__(self, width=640, height=480):
        self.width = width
        self.height = height
        self.bar_width = 0.8
        self.data_colour = '#f00'
        self.data_colours = ['#ed4334','#27ca42','#2546eb','#eadc45','#faab43']
        self.background_colour = '#efefef'

        self.axis_font_size = 24
        self.axis_stroke_colour = '#000'
        self.axis_stroke_width = '2'

        self.show_grid_lines = True 
        self.grid_lines_stroke_colour = '#a7a7a7'
        self.grid_lines_stroke_width = 1
        
        self.x_axis_interval = 1 # Show every n items on the x axis
        self.y_axis_tick_number = 7

        self.pie_inner_radius = 0.5 # Move out to separate function in future
        self.pie_value_radius = 1.1 # Move out to separate function in future

        self.text_styles = {
            'x_axis': TextStyle('x_axis'),
            'y_axis': TextStyle('y_axis'),
            'value': TextStyle('value'),
            'circle_value': CircleTextStyle('circle_value')
        }
        self.show_values = 'none'

        # L U R D
        self.margin = (0.05, 0.05, 0.05, 0.1)

    def __iadd__(self, other):
        if isinstance(other, TextStyle):
            if other.text_type:
                self.text_styles[other.text_type] = other
            else:
                print(f'Invalid text_type: TextStyle({other.text_type})')
                raise TypeError
        else:
            raise TypeError

        return self

    @property
    def css(self):
        css = ''

        for text_type in self.text_styles:
            css += self.text_styles[text_type].css

        css += f'''
        .data_colour {{
            fill: {self.data_colour};
        }}
        
        .line_chart_data_line {{
            stroke-width: 2px;
        }}
        
        .axis {{
            stroke: {self.axis_stroke_colour};
            stroke-width: {self.axis_stroke_width}px;
        }}

        .grid_lines {{
            stroke: {self.grid_lines_stroke_colour};
            stroke-width: {self.grid_lines_stroke_width}px;
        }}
        '''

        return css
